# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mock import patch

from charmworld.testing import JobTestBase

from charmworld.jobs.lp import available_charms
from charmworld.jobs.lp import queue_charms
from charmworld.jobs.utils import get_queue


def available_charms_mock(charm_data=None, limit=None):
    yield {'branch_dir': 'foo', 'branch_spec': 'bar'}


class TestLPJob(JobTestBase):

    def test_available_charms_parse_failure(self):
        # If the charm cannot be parsed, it is not returned.
        handler = self.get_handler("charm.launchpad")
        charm_data = [[u'~charmers/charms/precise/junk/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        charms = [charm for charm in available_charms(charm_data)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Unable to parse repo "
                   "~charmers/charms/precise/junk/someproject/trunk")
        self.assertIn(err_msg, log_messages)
        self.assertEqual([], charms)

    def test_available_charms_filter(self):
        # If the branch doesn't start with the defined filter (i.e.
        # ~charmers/charms/precise) it is not returned
        charm_data = [[u'~not-charmers/charms/precise/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        with patch('charmworld.jobs.lp.CHARM_IMPORT_FILTER',
                   '~charmers/charms/precise'):
            charms = [charm for charm in available_charms(charm_data)]
        self.assertEqual([], charms)

    def test_available_charms_not_trunk(self):
        # If the branch is not trunk or trunk-1, the charm is not returned by
        # available charms.
        handler = self.get_handler("charm.launchpad")
        charm_data = [[u'~charmers/charms/precise/someproject/foo',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]]
        charms = [charm for charm in available_charms(charm_data)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Skipped branch ~charmers/charms/precise/someproject/foo")
        self.assertIn(err_msg, log_messages)
        self.assertEqual([], charms)

    def test_available_charm_limit(self):
        handler = self.get_handler("charm.launchpad")
        charm_data = [[u'~charmers/charms/precise/someproject/trunk',
                       u'ja@appflower.com-20120329093714-s2m9e28dwotmijqc',
                       []]] * 200
        charms = [charm for charm in available_charms(charm_data, limit=110)]
        log_messages = [record.getMessage() for record in handler.buffer]
        err_msg = ("Import limit reached (110)... stopping")
        self.assertIn(err_msg, log_messages)
        self.assertEqual(110, len(charms))

    @patch('charmworld.jobs.lp.available_charms', available_charms_mock)
    def test_queue_charms(self):
        out_queue = get_queue('test_queue')
        self.addCleanup(out_queue.clear)
        queue_charms(out_queue=out_queue)
        item = out_queue.next()
        self.assertEqual(
            {'branch_dir': 'foo', 'branch_spec': 'bar'}, item.payload)
