# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from mongoqueue import MongoQueue
from mock import create_autospec

from charmworld.jobs.worker import CharmWorker
from charmworld.jobs.worker import DEFAULT_JOBS
from charmworld.jobs.ingest import IngestJob
from charmworld.testing import MongoTestBase


class SuccessfulRun(IngestJob):

    def run(self, charm_data):
        charm_data['ingested'] = True


class FailureRun(IngestJob):

    def run(self, charm_data):
        if charm_data['charm'] == 'foo':
            raise Exception('Exception forced by job.')
        else:
            charm_data['ingested'] = True


class GatheringRun(IngestJob):

    def setup(self):
        self.charms = []

    def run(self, charm_data):
        self.charms.append(charm_data)


class WorkerTest(MongoTestBase):

    def setUp(self):
        super(WorkerTest, self).setUp()
        self.queue = MongoQueue(self.db.queue, 'queue')
        self.errors = MongoQueue(self.db.error_queue, 'errors')
        self.queue.put({'charm': 'foo'})
        self.queue.put({'charm': 'bar'})

    def tearDown(self):
        self.queue.clear()
        self.errors.clear()
        super(WorkerTest, self).tearDown()

    def test_basic_running(self):
        # Barring errors, the worker will go through its entire set of
        # ingest jobs for a charm.
        ingest_jobs = (
            SuccessfulRun(),
            GatheringRun(),
        )
        worker = CharmWorker(ingest_jobs, self.queue, self.errors)
        worker.run(quit_on_empty=True)

        self.assertEqual(0, self.queue.size())
        self.assertEqual(0, self.errors.size())
        expected = [
            {'ingested': True, 'charm': 'foo'},
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = ingest_jobs[1].charms
        self.assertEqual(expected, charms)

    def test_failure_while_running(self):
        ingest_jobs = (FailureRun(), GatheringRun())
        worker = CharmWorker(ingest_jobs, self.queue, self.errors)
        worker.run(quit_on_empty=True)

        self.assertEqual(0, self.queue.size())
        self.assertEqual(1, self.errors.size())
        expected = [
            {'ingested': True, 'charm': 'bar'},
        ]
        charms = ingest_jobs[1].charms
        self.assertEqual(expected, charms)

    def test_stop_when_next_item_none(self):

        class LockedJobQueue:
            """Queue representing the state when a locked jobs exists."""

            @staticmethod
            def next():
                """A locked job cannot be processed."""
                return None

            @staticmethod
            def size():
                """If a queue has a locked job, its length > 0"""
                return 1

        ingest_jobs = (
            SuccessfulRun(),
            GatheringRun(),
        )
        worker = CharmWorker(ingest_jobs, LockedJobQueue, self.errors)
        worker.run(quit_on_empty=True)

    def test_ingest_job_contract(self):
        # All ingest jobs accept the right inputs.
        ingest_jobs = [create_autospec(job)() for job in DEFAULT_JOBS]
        worker = CharmWorker(ingest_jobs, self.queue, self.errors)
        worker.run(quit_on_empty=True)
        self.assertEqual(0, self.queue.size())
        self.assertEqual(0, self.errors.size())

    def test_interval_is_int(self):
        worker = CharmWorker([], self.queue, self.errors)
        self.assertIs(int, type(worker.interval),
                      'CharmWorker.interval is a %s, not an int.' %
                      type(worker.interval))
