# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import argparse
import logging
import time

from charmworld.utils import (
    configure_logging,
    get_ini,
)
from charmworld.jobs.config import CHARM_QUEUE
from charmworld.jobs.config import CHARM_ERROR_QUEUE
from charmworld.jobs.ingest import BzrIngestJob
from charmworld.jobs.ingest import ChangelogIngestJob
from charmworld.jobs.ingest import IndexIngestJob
from charmworld.jobs.ingest import JenkinsIngestJob
from charmworld.jobs.ingest import ProofIngestJob
from charmworld.jobs.ingest import StoreIngestJob
from charmworld.jobs.ingest import ScanIngestJob
from charmworld.jobs.utils import get_queue


DEFAULT_JOBS = (
    BzrIngestJob,
    StoreIngestJob,
    ProofIngestJob,
    JenkinsIngestJob,
    ChangelogIngestJob,
    ScanIngestJob,
    IndexIngestJob,
)

settings = get_ini()
INTERVAL = int(settings.get('worker_interval'))


class CharmWorker(object):
    """Processes raw charm data"""

    def __init__(self, ingest_jobs, queue, error_queue, interval=INTERVAL):
        self.ingest_jobs = ingest_jobs
        self.queue = queue
        self.errors = error_queue
        self.log = logging.getLogger('charm.worker')
        self.interval = interval

    def run(self, quit_on_empty=False):
        self.log.info('Starting processing')
        for job in self.ingest_jobs:
            job.setup()
        while True:
            item = self.queue.next()
            if item is None:
                if quit_on_empty:
                    self.log.info(
                        'Empty queue with quit_on_empty set. Quitting.')
                    break
                else:
                    wait_time_in_minutes = self.interval / 60
                    self.log.info(
                        'Empty queue. Waiting %s minutes to recheck.' %
                        wait_time_in_minutes)
                    time.sleep(self.interval)
                    continue
            charm_data = item.payload
            for job in self.ingest_jobs:
                stage = job.name
                log = logging.getLogger("charm.%s" % stage)
                try:
                    job.run(charm_data)
                except Exception as e:
                    err_msg = "%s error on %s: %s" % (
                        stage, charm_data, str(e))
                    log.exception(err_msg)
                    charm_error = {'error_stage': stage, 'error': str(e)}
                    if hasattr(e, 'output'):
                        charm_error['output'] = str(e.output)
                    charm_data['error'] = charm_error
                    self.errors.put(charm_data)
                    break
                finally:
                    item.complete()


# XXX j.c.sackett Feb 21 2013 Bug:1111708
# Functions not currently bound to jobs; probably need new jobs created for
# them outside of ingest worker.
#def reindex(db, indexer):
    #log = logging.getLogger("charm.index")
    #count = 0
    #for charm in db.find():
        #log.debug("Indexing %s", charm['branch_spec'])
        #try:
            #index_charm(indexer, charm)
        #except:
            #log.error("Indexing charm %s", charm['branch_spec'])
            #raise
        #count += 1
    #log.info("Indexed %d Charms" % count)


if __name__ == "__main__":
    configure_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--quit-on-empty",
        help="Shut down the worker if the queue is empty",
        action="store_true")
    args = parser.parse_args()
    queue = get_queue(CHARM_QUEUE)
    errors = get_queue(CHARM_ERROR_QUEUE)
    jobs = [job() for job in DEFAULT_JOBS]
    worker = CharmWorker(jobs, queue, errors)
    worker.run(quit_on_empty=args.quit_on_empty)
