# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

import gridfs
from os import listdir
from os.path import dirname
from os.path import isdir
from os.path import join
import pymongo
import random
import re

from charmworld.utils import quote_key


def getconnection(settings, fsync=False, safe=False):
    """Initiliaze the PyMongo database Connection object.

    :param settings: dict of application settings including the mongo.X info.

    """
    url_or_host = settings.get('mongo.url')
    if not url_or_host:
        url_or_host = settings.get('mongo.host')
        port = int(settings.get('mongo.port'))
    else:
        port = None
    connection = pymongo.Connection(
        url_or_host, port,
        fsync=fsync,
        safe=safe,
    )
    return connection


def getdb(connection, database_name):
    """Helper to access the db instance.

    This is needed because the test runner nose loads views which tried to
    import db directly, before the main() is run and initialize_db is run with
    the connection information.

    :param connection: pymongo connection object.
    :param database_name: mongo db name to connect to.
    """
    return getattr(connection, database_name)


def getfs(database):
    """Helper to get the gridfs handle used to store files."""
    return gridfs.GridFS(database)


def parse_sso_fields(payload):
    """Process the data form a sso request to a normal clean dict."""
    user_sso_data = set([
        'openid.sreg.nickname',
        'openid.sreg.fullname',
        'openid.sreg.email',
        'openid.sreg.timezone',
    ])
    common = user_sso_data.intersection(payload)
    # We're using the last bit of the key in our User object.
    user_data = dict((key.split('.')[2], payload[key]) for key in common)
    groups = []
    if payload['openid.lp.is_member'] != '':
        groups = payload['openid.lp.is_member'].split(',')
    user_data['groups'] = groups
    return user_data


class UserMgr(object):
    """Handle non-instance user actions."""

    @staticmethod
    def auth_principalfinder(userid, request):
        """Pyramid wants to know what principals are associated with a user.

        This is a helper provided for the RequestWithUser factory.
        """
        user = request.user
        if user is None:
            return None
        group_principals = ['group %s' % group for group in user.groups]
        return ['user %s' % user.userid] + group_principals

    @staticmethod
    def find_one(db, userid):
        found = db.users.find_one({'userid': userid})

        if not found:
            return None
        else:
            # Make sure we return a instantiated object vs the raw dictionary
            # from mongo.
            userid = found['userid']
            del(found['userid'])
            return User(userid, **found)

    @staticmethod
    def from_sso(payload):
        """Generate a User object instance from the SSO payload.

        We take the SSO dict and turn it into something that a User object,
        without being SSO aware, can easily handle and process.

        """
        user_data = parse_sso_fields(payload)
        user_id = payload['openid.claimed_id']
        return User(user_id, **user_data)


class User(object):
    """A user in our system."""

    # Attributes are defined as a list so we can easily create a
    # dict(someuser) helper to dump to mongodb.
    attributes = [
        '_id',
        'userid',
        'nickname',
        'fullname',
        'email',
        'groups',
        'timezone',
    ]

    def __init__(self, userid, **kwargs):
        self.userid = userid

        # Set some sane defaults since this depends on the users providing us
        # with the data.
        self.nickname = kwargs.get('nickname', 'Not Provided')
        self.fullname = kwargs.get('fullname', 'Not Provided')
        self.email = kwargs.get('email', None)
        self.groups = kwargs['groups']
        self.timezone = kwargs.get('timezone', 'GMT')

        # We need to worry about if this is a mongodb record or just a new
        # user object. A new object has no mongo id.
        if '_id' in kwargs:
            self._id = kwargs['_id']
        else:
            self._id = None

    def __iter__(self):
        """Allow the User to be dumped to a dict for mongo."""
        for key in self.attributes:
            yield(key, getattr(self, key))

    def in_group(self, group):
        return True if group in self.groups else False

    def save(self, db):
        """Actual storing of the instance into mongodb."""
        mongo_data = dict(self)
        if not self._id:
            db.users.save(mongo_data)
        else:
            self._id = db.users.insert(mongo_data)

    def update_from_sso(self, payload):
        """Update the stored user data from a fresh sso payload dump."""
        user_data = parse_sso_fields(payload)
        for key, val in user_data.items():
            setattr(self, key, val)


class QAData(object):
    """Build a usable object out of the QA questions and instance data."""

    def _load_questionset(self):
        """Load the qa questions we want to have data about."""
        stored_questions = self.db.qa.find()
        # We index them for easy fetching of each categories data later.
        self.qa_categories = dict([(q['name'], q) for q in stored_questions])

    def __init__(self, db, charm_data=None):
        self.db = db
        self._load_questionset()
        self.charm_data = charm_data

    @property
    def categories(self):
        return sorted(self.qa_categories)

    @property
    def score(self):
        """Calculate the score for whole qa set across categories."""

        # If we have no qa data then we need to respond with not available
        if not self.charm_data:
            return None

        score = 0
        for category in self.categories:
            category_score = self.category_score(category)
            if category_score is not None:
                score = score + category_score

        return score

    def category_score(self, category):
        """Given a category retrieve the scores for it."""
        score = 0
        if self.charm_data is None or category not in self.charm_data:
            return None

        category_questions = self.qa_categories[category]['questions']
        charm_category_scores = self.charm_data[category]

        for qnum, question in enumerate(category_questions):
            score_key = "{0}_{1}".format(category, qnum)
            q_score = charm_category_scores.get(score_key, 0)
            # The score could be '' if it's unknown vs yes/no
            if q_score:
                score = score + int(q_score)
        return score

    def category_max_score(self, category):
        """How many possible points does the category have?"""
        return len(self.qa_categories[category]['questions'])

    def filtered_scores(self):
        """Return only scores that match the categories."""
        if self.charm_data is None:
            return None
        score_categories = {}
        for category in self.qa_categories:
            if category not in self.charm_data:
                continue
            score_categories[category] = {}
            for question in self.qa_categories[category]['questions']:
                qid = question['id']
                score = self.charm_data[category].get(qid)
                if score is not None:
                    score_categories[category][qid] = score
        return score_categories


class CharmFileSet(object):
    """Handle Charm file storage and access.

    Charm key - series/owner/name/bname [branch name: trunk]

    """
    store_files = [
        'config.yaml',
        'metadata.yaml',
        'readme.*',
        'revision',
    ]

    store_directories = [
        'hooks',
    ]

    @classmethod
    def make_charm_file(cls, fs, charm_data, file_path, filename):
        charmid = cls.gen_charmid(charm_data)
        fileid = cls.gen_fileid(charm_data, file_path)
        return CharmFile(
            fs,
            fileid,
            charmid,
            filename=filename,
        )

    @classmethod
    def save_file(cls, fs, charm_data, filename, local_directory, subdir=None):
        """Do the actual storage of the file contents.

        :param fs: The mongo fs connection to use.
        :param fileid: The id of the file which is a full path and escaped.
        :param filename: The original short filename of the file.
        :param directory: The directory is this file located in on disk.

        """

        if subdir:
            complete_filename = join(subdir, filename)
        else:
            complete_filename = filename
        charm_file = cls.make_charm_file(
            fs, charm_data, complete_filename, filename)

        file_path = join(local_directory, complete_filename)
        with open(file_path) as handle:
            charm_file.save(handle.read())
        return charm_file

    @classmethod
    def gen_fileid(cls, charm_data, filename):
        """Generate our own mongo _id for the files.

        In keeping with the bzr structure files are stored:
        series-owner-name-bname-filename
        """
        return quote_key("/".join([
            cls.gen_charmid(charm_data),
            filename
        ]))

    @staticmethod
    def gen_charmid(charm_data):
        """Generate the charm id to aid in searching."""
        return "/".join([
            charm_data['owner'],
            charm_data['series'],
            charm_data['name'],
            charm_data['bname'],
        ])

    @staticmethod
    def get_by_id(fs, fileid):
        """Fetch the contents of the file

        :param fileid: The stored _id of the file.

        """
        if '.' in fileid:
            fileid = quote_key(fileid)
        found = fs.get(fileid)
        cf = CharmFile(fs, fileid, found.charmid, instance=found)
        return cf

    @classmethod
    def save_files(cls, fs, charm_data, charm_bzr_path):
        """Given a bzr branch, load the files we care about."""
        root_files = listdir(charm_bzr_path)
        stored = []

        for pattern in cls.store_files:
            # Make sure our file exists and find the right filename for it.
            found_filename = None
            for filename in root_files:
                m = re.match(pattern, filename, flags=re.I)
                if m:
                    found_filename = filename

            if found_filename:
                stored.append(
                    cls.save_file(
                        fs,
                        charm_data,
                        found_filename,
                        charm_bzr_path))

        for directory in cls.store_directories:
            dir_path = join(charm_bzr_path, directory)
            dir_files = listdir(dir_path)

            for filename in dir_files:
                if not isdir(join(dir_path, filename)):
                    stored.append(
                        cls.save_file(
                            fs, charm_data, filename, charm_bzr_path,
                            subdir=directory))
        return stored


class CharmFile(object):
    """Representation of the GridFS stored document.

    Wraps a PyMongo GridOut object instance.

    """
    # Attributes are defined as a list so we can easily create a
    # dict(charmfile) helper to dump to mongodb.
    attributes = [
        'fileid',
        'charmid',
        'subdir',
        'filename',
        'length',
        'md5',
    ]

    def __iter__(self):
        """Allow the Object to be dumped to a dict for mongo."""
        for key in self.attributes:
            if hasattr(self, key):
                yield(key, getattr(self, key))
            else:
                yield(key, getattr(self._gridout, key))

    @property
    def subdir(self):
        """From a file id drop the file and return a charmid"""
        filename = self.fileid.replace(quote_key(self.charmid), '').lstrip('/')
        # If there's no subdir in the filename, there's no subdir to return.
        return dirname(filename)

    @property
    def fileid(self):
        """fileid is more clear in code, but we're mapping this to mongo _id

        fileid cannot have . characters; it's used as a key in mongo.
        """
        if '.' in self._id:
            return quote_key(self._id)
        else:
            return self._id

    @property
    def filename(self):
        """The filename is in the fileid so auto pull it out."""
        if self._filename:
            return self._filename
        elif self._gridout:
            return self._gridout.filename
        else:
            return self._id.split('/')[-1]

    def __init__(self, fs, fileid, charmid, filename=None,
                 instance=None, metadata=None):
        """Get a file object started.

        :param fs: the gridfs client to use.
        :param fileid: The full mongo fileid to use for this file.
        :param charmid: The path of the charm identification through branch
            name.
        :param filename: Override the name of the file from being determined
            by the fileid.
        :param instance: The GridOut instance to be associated with the file.
        :param metadata: dict of extra data to store with the file in GridFS.

        """
        self.fs = fs
        self._id = fileid
        self.charmid = charmid
        self._filename = None

        if instance:
            self._gridout = instance
            # The filename might have been written out previously.
            self._filename = instance.filename
        else:
            self._gridout = None

        if filename:
            self._filename = filename

        if metadata:
            self.metadata = metadata
        else:
            self.metadata = {}

    def read(self):
        """Provide a read interface for the file."""
        if not self._gridout:
            self._gridout = self.fs.get(self.fileid)
        return self._gridout.read()

    def save(self, contents, data=None):
        """Store the contents of the file.

        :param contents: String/readable value of file content.
        :param data: dict of metadata to override defaults. See GridOut

        """
        if data:
            self.metadata.update(data)

        if self.fs.exists(self.fileid):
            self.fs.delete(self.fileid)
        self.metadata['_id'] = self.fileid
        self.metadata['charmid'] = self.charmid

        self.fs.put(
            contents,
            filename=self.filename,
            **self.metadata
        )

        # We then have to pull the file we just saved back out so that we can
        # load the instance data into our object.
        self._gridout = self.fs.get(self.fileid)
        return self.fileid


def acquire_session_secret(database):
    """Return a secret to use for AuthTkt sessions.

    If a secret is already in the database, use that.  Otherwise, generate a
    new secret.
    """
    secret = str(random.getrandbits(32))
    try:
        database.app_config.insert({'_id': 'session-secret', 'secret': secret})
    except pymongo.errors.DuplicateKeyError:
        # If there was a secret already, just use the existing one.
        pass
    unicode_secret = database.app_config.find_one('session-secret')['secret']
    return unicode_secret.encode('ascii')
