# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

import re

import pyelasticsearch
from pyelasticsearch.exceptions import ElasticHttpNotFoundError

from charmworld.utils import Timer

# Where wildcards are used, the final var name must be used, to avoid
# selecting any fields that may be integers, such as config.options.*.default,
# provides.*.interface.limit, or changes.revno.  Selecting an integer field
# causes a SearchParseException/NumberFormatException.
exact_fields = ['owner', 'series', 'provides.*.interface',
                'requires.*.interface']
free_text_fields = {
    'name': 10,
    'summary': 5,
    'description': 3,
    'config.options.*.description': None,
    'relations': None,
}


class InvalidCharmType(Exception):
    """Raised when an invalid charm type is specified."""


class NegativeLimit(Exception):
    """Raised when a negative limit is specified."""


class ElasticSearchClient:

    def __init__(self, client, index_name):
        self._client = client
        self.index_name = index_name
        self.ensure_index()

    def ensure_index(self):
        try:
            self._client.update_settings(self.index_name, {})
        except ElasticHttpNotFoundError:
            self._client.create_index(self.index_name)

    def wait_for_startup(self):
        self._client.health(self.index_name, wait_for_status='yellow')

    def delete_index(self):
        self._client.delete_index(self.index_name)

    @classmethod
    def from_settings(cls, settings, index_name='charms'):
        urls = settings['es_urls'].split(' ')
        return cls(pyelasticsearch.ElasticSearch(urls), index_name)

    @staticmethod
    def escape_query(query):
        return re.sub(r'[+\-&|!(){}\[\]\^"~*?:\\/]',
                      lambda m: r'\%s' % m.group(0), query)

    def index_charm(self, charm):
        self._client.put_mapping(self.index_name, 'charm', {
            'charm': {
                'properties': {
                    'name': {
                        'type': 'string', 'index': 'not_analyzed'
                    },
                    'owner': {
                        'type': 'string', 'index': 'not_analyzed'
                    },
                    'i_provides': {
                        'type': 'string', 'index': 'not_analyzed'
                    },
                    'i_requires': {
                        'type': 'string', 'index': 'not_analyzed'
                    },
                    'series': {
                        'type': 'string', 'index': 'not_analyzed'
                    }
                }
            }
        })
        charm_id = charm['_id']
        self._client.index(self.index_name, 'charm', charm, charm_id,
                           refresh=True)

    @staticmethod
    def _get_text_query(text):
        boosted_fields = [field + ('' if boost is None else '^%d' % boost)
                          for field, boost in free_text_fields.items()]
        if text != '':
            dsl = {
                'query_string': {
                    'query': text,
                    'fields': boosted_fields + exact_fields,
                }
            }
        else:
            dsl = {'match_all': {}}
        return dsl

    @classmethod
    def _get_official_boost(cls, dsl):
        dsl = {
            'custom_filters_score': {
                'query': dsl,
                'filters': [
                    {
                        'filter': {
                            'term': {
                                'owner': 'charmers'
                            }
                        },
                        'boost': 10,
                    },
                ],
            }
        }
        return dsl

    @staticmethod
    def _get_filtered(dsl, filters, type_):
        and_clause = {
            'and': [
                {'terms': {key: value}} for key, value in filters.items()
            ]
        }
        if type_ is not None:
            typeclauses = []
            for value in type_:
                if value == 'approved':
                    typeclauses.append({'term': {'owner': 'charmers'}})
                elif value == 'community':
                    typeclauses.append({
                        'not': {
                            'term': {'owner': 'charmers'}
                        }
                    })
                elif value == 'environment':
                    typeclauses.append({'term': {'_id': ''}})
                else:
                    raise InvalidCharmType(value)

            if len(typeclauses) > 0:
                and_clause['and'].append({'or': typeclauses})

        if len(and_clause['and']) == 0:
            and_clause = {'match_all': {}}
        dsl = {
            'query': {
                'filtered': {
                    'filter': and_clause,
                    'query': dsl,
                }
            }
        }
        return dsl

    def api_search(self, text, filters, type_, limit=None):
        dsl = self._get_text_query(text)
        dsl = self._get_official_boost(dsl)
        dsl = self._get_filtered(dsl, filters, type_)
        if limit is None:
            status = self._client.status(index=self.index_name)['indices']
            limit = status[self.index_name]['docs']['num_docs']
        if limit < 0:
            raise NegativeLimit()
        hits = self._client.search(dsl, index=self.index_name, size=limit)
        return [hit['_source'] for hit in hits['hits']['hits']]

    def search(self, terms):
        size = None
        if terms == '':
            size = 0

        dsl = self._get_text_query(terms)
        dsl = {'query': self._get_official_boost(dsl)}

        with Timer() as timer:
            status = self._client.status(index=self.index_name)['indices']
            charm_total = status[self.index_name]['docs']['num_docs']
            if size is None:
                size = charm_total
            hits = self._client.search(dsl, index=self.index_name, size=size)
        charms = [{'data': hit['_source'], 'weight': hit['_score']}
                  for hit in hits['hits']['hits']]
        return {
            'results': charms,
            'search_time': u"%.5f" % timer.duration(),
            'charm_total': charm_total,
            'matches': len(charms),
            'matches_human_readable_estimate': '%d' % len(charms),
        }
