# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

"""Unit tests for our application models."""
from os.path import dirname
from os.path import join

from charmworld.models import acquire_session_secret
from charmworld.models import QAData
from charmworld.models import CharmFile
from charmworld.models import CharmFileSet
from charmworld.models import UserMgr
from charmworld.testing import factory
from charmworld.testing import MongoTestBase
from charmworld.utils import quote_key


class TestUser(MongoTestBase):
    """Unit tests for our User object wrapping of the mongo dict."""

    def test_model_from_sso(self):
        """Given an SSO dump we can get back a decent model object."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)

        self.assertEqual(payload['openid.claimed_id'], user.userid)
        self.assertEqual(payload['openid.sreg.nickname'], user.nickname)
        self.assertEqual(payload['openid.sreg.fullname'], user.fullname)
        self.assertEqual(payload['openid.sreg.email'], user.email)
        self.assertEqual(payload['openid.sreg.timezone'], user.timezone)
        self.assertEqual(payload['openid.lp.is_member'], ','.join(user.groups))

    def test_user_limited_sharing(self):
        """The user might limit information shared."""
        claimed_id = u'https://login.ubuntu.com/+id/MSbTCYM'

        user = UserMgr.from_sso({
            'openid.claimed_id': claimed_id,
            # When a user chooses not to share teams, an empty is_member is
            # returned.
            'openid.lp.is_member': '',
        })

        self.assertEqual(claimed_id, user.userid)
        self.assertEqual('Not Provided', user.nickname)
        self.assertEqual('Not Provided', user.fullname)
        self.assertEqual(None, user.email)
        self.assertEqual('GMT', user.timezone)
        self.assertEqual([], user.groups)

    def test_can_dict_user(self):
        """We need to be able to turn User into a dict for mongo."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)

        dict_user = dict(user)
        self.assertEqual(
            payload['openid.claimed_id'],
            dict_user['userid'])
        self.assertEqual(
            payload['openid.sreg.nickname'],
            dict_user['nickname'])

    def test_user_stored_in_mongo(self):
        """The user should store in mongo correctly on save."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        user.save(self.db)

        found = self.db.users.find_one({'userid': user.userid})
        self.assertEqual(user.userid, found['userid'])

    def test_user_from_mongo(self):
        """Pulling the user from mongo via UserMgr should get a User out."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        user.save(self.db)

        found = UserMgr.find_one(self.db, user.userid)
        self.assertEqual(user.userid, found.userid)

    def test_user_updated(self):
        """Saving will either insert or save as required."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        # Does the insert
        user.save(self.db)

        found = UserMgr.find_one(self.db, user.userid)
        self.assertEqual(user.userid, found.userid)
        self.assertEqual(user.fullname, found.fullname)

        found.fullname = 'Changed in test'
        found.save(self.db)

        found2 = UserMgr.find_one(self.db, found.userid)
        self.assertEqual(found.userid, found2.userid)
        self.assertEqual(found.fullname, found2.fullname)

    def test_update_from_sso(self):
        """A new set of SSO data updates the user found."""
        payload = factory.makeSSOResponse()
        user = UserMgr.from_sso(payload)
        # Does the insert
        user.save(self.db)

        new_groups = ['launchpad']
        payload['openid.lp.is_member'] = new_groups[0]
        user.update_from_sso(payload)
        self.assertEqual(new_groups, user.groups)

    def test_auth_principalfinder(self):
        class Request:
            user = UserMgr.from_sso(
                factory.makeSSOResponse(groups=['charmers', 'canonical']))
        expected_principals = [
            'user https://login.ubuntu.com/+id/MSbTCYM',
            'group charmers',
            'group canonical']
        principals = UserMgr.auth_principalfinder('', Request)
        self.assertEqual(expected_principals, principals)


class TestQASet(MongoTestBase):
    """Unit tests for our User object wrapping of the mongo dict."""

    def setUp(self):
        super(TestQASet, self).setUp()
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)

    def test_category_property(self):
        qadata = QAData(self.db)
        self.assertEqual(['reliable'], qadata.categories)

    def test_none_overall_if_no_data(self):
        """When no charm qa data we get None to indicate N/A."""
        qadata = QAData(self.db)
        self.assertIsNone(qadata.score)

    def test_overall_score(self):
        scores = factory.makeQAScore()
        qadata = QAData(self.db, charm_data=scores)

        self.assertEqual(4, qadata.score)

    def test_none_category_score_if_missing_data(self):
        """When no scores are available for the cateogry we get None

        None is our indicator of a score that's N/A

        """
        qadata = QAData(self.db)
        self.assertIsNone(qadata.category_score('reliable'))

    def test_category_score(self):
        scores = factory.makeQAScore()
        qadata = QAData(self.db, charm_data=scores)

        self.assertEqual(4, qadata.category_score('reliable'))

    def test_category_max_score(self):
        """Determine the total available points for the category."""
        qadata = QAData(self.db)
        self.assertEqual(9, qadata.category_max_score('reliable'))


class TestCharmFileSet(MongoTestBase):
    # Escape the base_fileid since all code will escape this. We're comparing
    # against this string so it needs to also be escaped.
    base_fileid = 'charmers/precise/charmworld%2Etest/trunk/'
    charm_data = {
        'series': 'precise',
        'owner': 'charmers',
        'name': 'charmworld.test',
        'bname': 'trunk',
    }
    charm_path = join(
        dirname(__file__),
        'testing/data/sample_charm'
    )

    def test_fileid_generation(self):
        """A charm must generate an unique fileid."""
        self.assertEqual(
            'charmers/precise/charmworld%2Etest/trunk/readme',
            CharmFileSet.gen_fileid(self.charm_data, 'readme')
        )

    def test_load_bzr_files(self):
        """Processing a charm dir loads the file contents."""
        CharmFileSet.save_files(
            self.fs,
            self.charm_data,
            self.charm_path)

        # There should be the following files stored into the gridfs
        # filestore.
        expected_files = [
            quote_key('config.yaml'),
            quote_key('metadata.yaml'),
            quote_key('README.md'),
            'hooks/config-changed',
            'hooks/database-relation-broken',
            'hooks/database-relation-changed',
            'hooks/install',
            'hooks/linkstart',
            'hooks/restart',
            'hooks/start',
            'hooks/stop',
            'hooks/upgrade-charm',
            'hooks/website-relation-changed',
        ]

        for name in expected_files:
            fileid = self.base_fileid + name
            self.assertTrue(
                self.fs.exists(fileid),
                'Should find file:' + fileid)

    def test_case_kept_for_filenames(self):
        """README.md is lower cased in the key, but upper in filename."""
        CharmFileSet.save_files(
            self.fs,
            self.charm_data,
            self.charm_path)

        # Verify that the filename of the README is kept.
        found = self.fs.get_last_version('README.md')
        # fileid is escaped because it's used as a mongo key.
        expected = self.base_fileid + quote_key('README.md')
        self.assertEqual(expected, found._id)

    def test_get_by_id(self):
        """Load a CharmFile instance from the datastore."""
        CharmFileSet.save_files(
            self.fs,
            self.charm_data,
            self.charm_path)

        readme_id = self.base_fileid + quote_key('README.md')

        cf = CharmFileSet.get_by_id(self.fs, readme_id)
        self.assertEqual('README.md', cf.filename)
        self.assertEqual(readme_id, cf.fileid)

        # We need to be able to dict the metadata to store in the charm record
        # itself.
        metadata = dict(cf)
        self.assertEqual('README.md', metadata['filename'])
        self.assertEqual(readme_id, metadata['fileid'])
        self.assertEqual(u'charmers/precise/charmworld.test/trunk',
                         metadata['charmid'])
        self.assertEqual(2757, metadata['length'])
        self.assertEqual(u'24f0bade2fb27e7b3b4277f1c930fd11', metadata['md5'])
        self.assertEqual('', metadata['subdir'])

    def test_get_by_id_hooks_file(self):
        """The hooks file is a subdir down and we want to track that."""
        CharmFileSet.save_files(
            self.fs,
            self.charm_data,
            self.charm_path)

        install_id = self.base_fileid + 'hooks/install'

        cf = CharmFileSet.get_by_id(self.fs, install_id)
        self.assertEqual('install', cf.filename)
        self.assertEqual('hooks', cf.subdir)
        self.assertEqual(install_id, cf.fileid)

        # We need to be able to dict the metadata to store in the charm record
        # itself.
        metadata = dict(cf)
        self.assertEqual('hooks', metadata['subdir'])


class TestCharmFile(MongoTestBase):

    def test_filename_parsing(self):
        """The filename is the end of the fileid."""
        cf = CharmFile(
            self.fs, '/test/file/path.txt', '/charmers/precise/redis')
        self.assertEqual('path.txt', cf.filename)

    def test_filename_with_caps(self):
        """Whe can specifiy the filename to keep it original."""

        cf = CharmFile(
            self.fs, '/test/file/path.txt', '/charmers/precise/redis',
            filename='PATH.txt')
        self.assertEqual('PATH.txt', cf.filename)

    def test_storing_file_contents(self):
        contents = 'Contents of the file.'
        fileid = '/test/file/path.txt'
        cf = CharmFile(self.fs, fileid, '/charmers/precise/redis')
        cf.save(contents)

        # Make sure we check for the escaped fileid to see it exists.
        self.assertTrue(self.fs.exists(cf.fileid))

        found = self.fs.get(cf.fileid)
        self.assertEqual(contents, found.read())


class TestAcquireSessionSecret(MongoTestBase):

    def test_acquire_session_secret(self):
        """Multiple calls to acquire_session_secret return the same value."""
        secret = acquire_session_secret(self.db)
        self.assertIs(str, type(secret))
        self.assertEqual(secret, acquire_session_secret(self.db))

    def test_db_backed(self):
        """The secret is stored in the database.

        An existing session-secret is used if present.  Deleting it causes a
        new secret to be generated.
        """
        self.db.app_config.insert({'_id': 'session-secret',
                                   'secret': 'arbitrary-string'})
        first_secret = acquire_session_secret(self.db)
        self.assertEqual('arbitrary-string', first_secret)
        self.db.app_config.remove('session-secret')
        second_secret = acquire_session_secret(self.db)
        self.assertNotEqual(first_secret, second_secret)
