# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from pyelasticsearch.exceptions import ElasticHttpNotFoundError

from charmworld.jobs.ingest import make_store_url
from charmworld.search import (
    exact_fields,
    free_text_fields,
    InvalidCharmType,
    NegativeLimit,
)
from charmworld.testing import (
    factory,
    temp_index_client,
    TestCase,
)


class TestTempIndexClient(TestCase):

    def test_temp_index_client(self):
        with temp_index_client() as client:
            self.assertNotEqual('charms', client.index_name)
            client._client.update_settings(client.index_name, {})
        with self.assertRaises(ElasticHttpNotFoundError):
            client._client.update_settings(client.index_name, {})


class TestElasticSearchClient(TestCase):

    def setUp(self):
        super(TestElasticSearchClient, self).setUp()
        self.use_index_client()

    def exists_in_index(self, charm_id):
        client = self.index_client
        try:
            result = client._client.get(client.index_name, 'charm', charm_id)
        except ElasticHttpNotFoundError:
            return False
        return result['exists']

    def test_index_charm(self):
        foo_charm = factory.get_charm_json()
        self.index_client.wait_for_startup()
        self.assertFalse(self.exists_in_index(foo_charm['_id']))
        self.index_client.index_charm(foo_charm)
        self.assertTrue(self.exists_in_index(foo_charm['_id']))

    def makeCharm(self, *args, **kwargs):
        charm = factory.get_charm_json(*args, **kwargs)
        self.index_client.index_charm(charm)
        return charm

    def test_search(self):
        result_keys = ['results', 'search_time', 'charm_total', 'matches',
                       'matches_human_readable_estimate']
        client = self.index_client
        foo_charm = self.makeCharm()
        results = client.search('foo')
        self.assertItemsEqual(result_keys, results.keys())
        self.assertEqual([], results['results'])
        results = client.search(foo_charm['name'])['results']
        self.assertEqual(1, len(results))
        self.assertEqual(foo_charm, results[0]['data'])
        # The absolute value of weight is not defined, so we just ensure
        # it's present.
        self.assertItemsEqual(['data', 'weight'], results[0].keys())

    def search_ids(self, terms):
        client = self.index_client
        return [hit['data']['_id'] for hit in client.search(terms)['results']]

    def test_search_matches_on_search_terms(self):
        fields = free_text_fields.keys() + exact_fields
        fields = [field.split('.')[0] for field in fields]
        charm = factory.get_charm_json()
        # Ensure all fields have unique values.
        charm['owner'] = 'steve'
        charm['series'] = 'grumpy'
        charm['name'] = 'arthur'
        charm['label'] = 'organic'
        charm['bname'] = 'my-dev-branch'
        charm['i_provides'] = ['provisions']
        charm['provides']['website']['interface'] = 'gopher'
        charm['requires']['database']['interface'] = 'reiserfs'
        charm['_id'] = 'asdf'
        # Best-effort ensure there's no overlap between search terms
        items = charm.items()
        for num, (key, value) in enumerate(items):
            if not isinstance(value, basestring):
                continue
            for key2, value2 in items[num + 1:]:
                if not isinstance(value2, basestring):
                    continue
                self.assertNotIn(value, value2, '%s in %s' % (key, key2))
                self.assertNotIn(value2, value, '%s in %s' % (key2, key))
        self.index_client.index_charm(charm)
        for key, query in items:
            # Skip revision because it's an int.
            if key == 'revision':
                continue
            elif key in ('hooks', 'i_requires', 'i_provides'):
                query = query[0]
            elif key == 'changes':
                query = query[0]['message']
            elif key in ('provides', 'requires'):
                query = query.values()[0].values()[0]
            elif key == 'config':
                query = query.values()[0].values()[0]['description']
            elif key == 'store_data':
                query = query['digest']
            elif key == 'files':
                query = query['readme']['filename']
            elif key in ('first_change', 'last_change'):
                query = query['message']
            elif key == 'proof':
                query = query['w'][0]
            self.assertIsInstance(query, basestring)
            query = '"%s"' % self.index_client.escape_query(query)
            ids = self.search_ids(query)
            if key in fields:
                self.assertEqual(1, len(ids), 'No hit for %s' % key)
            else:
                self.assertEqual(0, len(ids), 'Undesired hit on %s' % key)

    def test_official_charms_boost(self):
        """An official version of a charm rates 10x higher."""
        charm = factory.get_charm_json()
        official_id = charm['_id']
        unofficial_id = official_id + '-1'
        client = self.index_client
        client.index_charm(charm)
        charm['owner'] = 'jrandom'
        charm['_id'] = unofficial_id
        client.index_charm(charm)
        results = client.search('terminal')['results']
        by_id = dict((result['data']['_id'], result) for result in results)
        self.assertAlmostEqual(by_id[official_id]['weight'],
                               by_id[unofficial_id]['weight'] * 10)

    def test_api_search_all(self):
        """With no parameters, all charms are returned."""
        one = self.makeCharm(name='name1')
        maintainer2 = 'jrandom@example.com (J. Random Hacker)'
        provides2 = {'cookie-factory': {
            'interface': 'imap',
        }}
        requires2 = {'postal-service': {
            'interface': 'envelope',
        }}
        options2 = {
            'foobar': {
                'type': 'str',
                'default': 'baz',
                'description': 'quxx',
            },
        }
        files2 = {
            'readme': {
                'filename': 'README.md',
                'subdir': '',
            },
            'config-changed': {
                'filename': 'config-changed',
                'subdir': 'hooks',
            },
        }
        two = self.makeCharm(owner='jrandom', revision=3,
                             series='warty', summary='Charm two',
                             name='name2', revno=13,
                             maintainer=maintainer2,
                             commit_message='message2',
                             provides=provides2,
                             requires=requires2, options=options2,
                             files=files2, subordinate=True)
        result = self.get_charms()
        self.assertEqual(2, len(result))
        one_url = make_store_url(1, charm=one)
        self.assertEqual(one, result[0])
        two_url = make_store_url(1, charm=two)
        self.assertEqual(two, result[1])
        self.assertNotEqual(one['_id'], two['_id'])
        self.assertNotEqual(one_url, two_url)

    def get_charms(self, text='', type_=None, *args, **kwargs):
        filters = dict((key, [value]) for key, value in kwargs.items())
        if type_ is not None and isinstance(type_, basestring):
            type_ = [type_]
        return self.index_client.api_search(text, filters, type_)

    def _test_filter(self, filter_str, make_charm=None):
        if make_charm is None:
            def make_charm(value, warty=False):
                kwargs = {filter_str: value}
                if filter_str != 'series' and warty:
                    kwargs['series'] = 'warty'
                return self.makeCharm(**kwargs)
        one = make_charm('common')
        self.index_client.index_charm(one)
        two = make_charm('common', warty=True)
        self.index_client.index_charm(two)
        three = make_charm('different')
        self.index_client.index_charm(three)
        four = make_charm('different-again')
        self.index_client.index_charm(four)
        result = self.get_charms(**{filter_str: 'common'})
        self.assertEqual(2, len(result))
        self.assertEqual(set([one['store_url'], two['store_url']]),
                         set(charm['store_url'] for charm in result))
        result = self.get_charms(**{filter_str: 'different'})
        self.assertEqual(set([three['store_url']]),
                         set(charm['store_url'] for charm in result))
        self.assertEqual(1, len(result))
        result = self.get_charms(**{filter_str: ['different', 'common']})
        self.assertEqual(3, len(result))

    def test_charms_name_filter(self):
        """Charms can be filtered by name."""
        self._test_filter('name')

    def test_charms_series_filter(self):
        """Charms can be filtered by series."""
        self._test_filter('series')

    def test_charms_owner_filter(self):
        """Charms can be filtered by owner."""
        self._test_filter('owner')

    def test_charms_provides_filter(self):
        """Charms can be filtered by provided interfaces."""
        def make_charm(value, warty=False):
            provides = {'foo': {'interface': value}}
            return self.makeCharm(provides=provides)
        self._test_filter('i_provides', make_charm)

    def test_charms_requires_filter(self):
        """Charms can be filtered by required interfaces."""
        def make_charm(value, warty=False):
            requires = {'foo': {'interface': value}}
            return self.makeCharm(requires=requires)
        self._test_filter('i_requires', make_charm)

    def test_charms_type_filter(self):
        """Filtering by type works as well as possible.

        Charms are considered 'approved' if owned by charmers, 'community'
        otherwise.  'environment' causes no charms to be returned.
        """
        self.makeCharm()
        self.makeCharm(owner='steve')
        result = self.get_charms(type_='community')
        self.assertEqual(['steve'], [charm['owner'] for charm in result])
        result = self.get_charms(type_='approved')
        self.assertEqual(['charmers'], [charm['owner'] for charm in result])
        result = self.get_charms(type_='environment')
        self.assertEqual([], [charm['owner'] for charm in result])
        with self.assertRaises(InvalidCharmType):
            self.get_charms(type_='foo')
        result = self.get_charms(type_=['approved', 'community'])
        self.assertEqual(['charmers', 'steve'],
                         [charm['owner'] for charm in result])

    def test_charms_text_search(self):
        abc_charm = self.makeCharm(summary='abc def')
        self.makeCharm(summary='ghi jkl')
        charms = self.get_charms(text='def')
        self.assertEqual(1, len(charms))
        self.assertEqual(abc_charm['_id'], charms[0]['_id'])

    def test_charms_text_search_respects_filters(self):
        official = self.makeCharm(summary='abc def')
        self.makeCharm()
        self.makeCharm(summary='abc def', owner='steve')
        self.makeCharm(owner='steve')
        result = self.get_charms(type_=['approved'], text='def')
        self.assertEqual([official['_id']],
                         [charm['_id'] for charm in result])

    def test_charms_limit(self):
        """Limit provides a maximum for charm numbers."""
        self.index_client.wait_for_startup()
        result = self.index_client.api_search('', {}, None, 6)
        self.assertEqual(0, len(result))
        self.makeCharm()
        self.makeCharm(owner='jrandom')
        result = self.index_client.api_search('', {}, None, 6)
        self.assertEqual(2, len(result))
        result = self.index_client.api_search('', {}, None, 1)
        self.assertEqual(1, len(result))
        result = self.index_client.api_search('', {}, None, 0)
        self.assertEqual(0, len(result))
        with self.assertRaises(NegativeLimit):
            result = self.index_client.api_search('', {}, None, -1)
