# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from contextlib import contextmanager
import logging
import os
from StringIO import StringIO
import unittest
from urllib2 import URLError

import mock
from mongoqueue import MongoQueue
from pyramid import testing
from webtest import TestApp

from charmworld import main
from charmworld.models import getconnection
from charmworld.models import getdb
from charmworld.models import getfs
from charmworld.search import ElasticSearchClient
from charmworld.testing.factory import makeSSOResponse
from charmworld.utils import get_ini


INI = get_ini()
MONGO_URL = os.environ.get('TEST_MONGODB')
MONGO_DATABASE = INI.get('mongo.database')
# Update the ini values with the database fired up by the nosemongo plugin.
INI['mongo.url'] = MONGO_URL

REAL_FILE_CHARM_DIR = os.path.join(os.path.split(__file__)[0], 'data')


class TestCase(unittest.TestCase):

    def get_handler(self, log_name):
        logger = logging.getLogger(log_name)
        handler = logging.handlers.BufferingHandler(20)
        logger.addHandler(handler)

        def cleanup():
            logger.removeHandler(handler)
        self.addCleanup(cleanup)
        return handler

    def use_context(self, context):
        result = context.__enter__()
        self.addCleanup(context.__exit__, None, None, None)
        return result

    def use_index_client(self):
        self.index_client = self.use_context(temp_index_client())
        return self.index_client


class MongoTestBase(TestCase):

    @property
    def connection(self):
        if getattr(self, '_connection', None) is None:
            self._connection = getconnection(INI, safe=True)
        return self._connection

    def setUp(self):
        self.db = getdb(self.connection, MONGO_DATABASE)
        self.fs = getfs(self.db)

    def tearDown(self):
        # Drop the Mongo database.
        self.connection.drop_database(MONGO_DATABASE)


class WebTestBase(MongoTestBase):

    def setUp(self):
        super(WebTestBase, self).setUp()
        app = main({}, **INI)
        self.app = TestApp(app)

    def tearDown(self):
        self.app = None
        super(WebTestBase, self).tearDown()


class ViewTestBase(MongoTestBase):

    def setUp(self):
        super(ViewTestBase, self).setUp()
        self.config = testing.setUp()
        self.index_client = None

    def getRequest(self, params=None, post=None):
        request = testing.DummyRequest(params=params, post=post)
        request.db = self.db
        request.index_client = self.index_client
        return request

    def tearDown(self):
        testing.tearDown()
        super(ViewTestBase, self).tearDown()


class RemoteURLBase(MongoTestBase):

    url_prefix = "https://jenkins.qa.ubuntu.com/job/"
    url_debug = False

    def setUp(self):
        super(RemoteURLBase, self).setUp()
        self.url_result_map = {"": ""}

    def lookup_url_result(self, url):
        if self.url_debug:
            print "url", url
        short_url = url[len(self.url_prefix):]
        if short_url in self.url_result_map:
            return StringIO(self.url_result_map[short_url])
        raise URLError(url)

    def invoke(self, func, *args, **kw):
        with mock.patch("urllib2.urlopen") as urlopen:
            urlopen.side_effect = self.lookup_url_result
            return func(*args, **kw)


class JobTestBase(MongoTestBase):

    def patch_ingest_job(self, job, methods):
        # Mocks multiple attr on a job.
        old_methods = []
        for m in methods:
            old_methods.append(getattr(job, m))
            setattr(job, m, mock.Mock())

        def cleanup():
            for i, m in enumerate(methods):
                setattr(job, m, old_methods[i])
        self.addCleanup(cleanup)

    def get_mongo_queues(self, data=None):
        in_queue = MongoQueue(self.db.in_queue, "in-queue")
        out_queue = MongoQueue(self.db.out_queue, "out-queue")

        if data:
            in_queue.put(data)

        def cleanup():
            in_queue.clear()
            out_queue.clear()
        self.addCleanup(cleanup)
        return (in_queue, out_queue)


@contextmanager
def login_request(app, groups=None):
    """Context manager to help functionally test a logged in user account.

    :param app: The application constructed for testing. Usually self.app when
                using WebTestBase.
    :return response: The post-logged in WebTest response object.
    """
    sso_resp = makeSSOResponse(groups=groups)
    resp = app.get(
        '/auth_callback',
        params=sso_resp,
        status=302)
    resp = resp.follow()
    yield resp


def get_bad_mock(exception_type=Exception, *args):
    """Convenience helper to generate a mock that will raise a particular
    exception with necessary messages.
    """
    if not len(args):
        args = ["Exception forced by mock."]
    return mock.Mock(side_effect=exception_type(*args))


@contextmanager
def temp_index_client():
    client = ElasticSearchClient.from_settings(INI, 'temp_index')
    try:
        yield client
    finally:
        client.delete_index()
