# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime
import hashlib

from pyramid.httpexceptions import HTTPNotFound
from webob.multidict import MultiDict

from charmworld.jobs.ingest import make_store_url
from charmworld.views.api import (
    API0,
    json_error,
)
from charmworld.testing import factory
from charmworld.testing import (
    ViewTestBase,
)


class TestJsonError(ViewTestBase):

    def test_json_error(self):
        """json_error produces expected response."""
        error = json_error(200, {'type': 'not_an_error'})
        self.assertEqual(200, error.status_code)
        self.assertEqual({'type': 'not_an_error'}, error.json_body)
        self.assertEqual('application/json', error.content_type)
        self.assertIn(("Access-Control-Allow-Origin", "*"), error.headerlist)
        self.assertIn(
            ("Access-Control-Allow-Headers", "X-Requested-With"),
            error.headerlist)


class APITestBase(ViewTestBase):

    def make_request(self, endpoint, remainder, **kwargs):
        """Make a request for a given endpoint, remainder and params."""
        request = self.getRequest()
        request.matchdict['endpoint'] = endpoint
        if remainder is not None:
            request.matchdict['remainder'] = remainder
        request.GET = MultiDict()
        for key, values in kwargs.iteritems():
            key = {'type_': 'type'}.get(key, key)
            if isinstance(values, basestring):
                values = [values]
            for value in values:
                request.GET.add(key, value)
        return request

    def get_response(self, endpoint, remainder='', **kwargs):
        """Get a Response for a given endpoint, remainder and params."""
        return API0(self.make_request(endpoint, remainder, **kwargs))()


class TestAPI0Charms(APITestBase):

    def setUp(self):
        super(ViewTestBase, self).setUp()
        self.use_index_client()

    def makeCharm(self, *args, **kwargs):
        charm_id, charm = factory.makeCharm(self.db, *args, **kwargs)
        self.index_client.index_charm(charm)
        return charm_id, charm

    def test_remote_friendly_headers(self):
        """The Juju-Gui needs to be able to query the api via diff origin."""
        self.index_client.wait_for_startup()
        resp = self.get_response('charms')
        self.assertIn(("Access-Control-Allow-Origin", "*"), resp.headerlist)
        self.assertIn(
            ("Access-Control-Allow-Headers", "X-Requested-With"),
            resp.headerlist)

    def test_charms(self):
        """Charms endpoint produces expected result."""
        one_changes = [factory.makeChange(created=datetime(1999, 12, 31),
                                          revno=37, message='Hello')]
        one_id, one = self.makeCharm(name='name1', changes=one_changes)
        self.index_client.index_charm(one)
        maintainer2 = 'jrandom@example.com (J. Random Hacker)'
        provides2 = {'cookie-factory': {
            'interface': 'imap',
        }}
        requires2 = {'postal-service': {
            'interface': 'envelope',
        }}
        options2 = {
            'foobar': {
                'type': 'str',
                'default': 'baz',
                'description': 'quxx',
            },
        }
        files2 = {
            'readme': {
                'filename': 'README.md',
                'subdir': '',
            },
            'config-changed': {
                'filename': 'config-changed',
                'subdir': 'hooks',
            },
        }
        two_changes = [factory.makeChange(created=datetime(1998, 11, 30),
                                          revno=32, message='Hi',
                                          authors=['Baz <baz@example.org>'])]
        two_id, two = self.makeCharm(owner='jrandom', revision=3,
                                     series='warty', summary='Charm two',
                                     name='name2', revno=13,
                                     maintainer=maintainer2,
                                     commit_message='message2',
                                     provides=provides2,
                                     requires=requires2, options=options2,
                                     files=files2, subordinate=True,
                                     changes=two_changes)
        response = self.get_response('charms')
        self.assertEqual(200, response.status_code)
        self.assertEqual('application/json', response.content_type)
        result = response.json_body['result']
        self.assertEqual(2, len(result))
        one_url = make_store_url(1, charm=one)
        self.assertEqual({'owner': 'charmers',
                          'maintainer': {
                              'name': 'John Doe',
                              'email': 'jdoe@example.com',
                          },
                          'id': one_url[3:],
                          'date_created': 0,
                          'name': 'name1',
                          'url': one_url,
                          'description': one['description'],
                          'revision': 1,
                          'summary': "Remote terminal classroom over https",
                          'distro_series': 'precise',
                          'is_approved': True,
                          'rating_numerator': 0,
                          'rating_denominator': 0,
                          'is_new': False,
                          'is_popular': False,
                          'code_source': {
                              'type': 'bzr',
                              'location':
                                  'lp:~charmers/charms/precise/name1/trunk',
                              'revision': '12',
                              'last_log': 'maintainer',
                              'bugs_link':
                                  'https://bugs.launchpad.net/charms/'
                                  'precise/+source/name1',
                              'revisions': [{
                                  'revno': 37,
                                  'message': 'Hello',
                                  'date': '1999-12-31T00:00:00',
                                  'authors': [
                                      {
                                          'name': 'Foo',
                                          'email': 'foo@example.com'
                                      },
                                      {
                                          'name': 'Bar',
                                          'email': 'bar@example.com'
                                      },
                                  ],
                              }],
                          },
                          'relations': {
                              'provides': {
                                  'website': {
                                      'interface': 'https',
                                  },
                              },
                              'requires': {
                                  'database': {
                                      'interface': 'mongodb',
                                  },
                              },
                          },
                          'options': {
                              'script-interval': {
                                  'type': 'int',
                                  'default': 5,
                                  'description':
                                  'The interval between script runs.'
                              },
                          },
                          'files': ['README', 'hooks/install'],
                          'tested_providers': [],
                          'downloads_in_past_30_days': 0,
                          'is_subordinate': False,
                          }, result[0])
        two_url = make_store_url(1, charm=two)
        self.assertEqual({'owner': 'jrandom',
                          'maintainer': {
                              'name': 'J. Random Hacker',
                              'email': 'jrandom@example.com',
                          },
                          'id': two_url[3:],
                          'date_created': 0,
                          'name': 'name2',
                          'url': two_url,
                          'description': two['description'],
                          'revision': 3,
                          'summary': 'Charm two',
                          'distro_series': 'warty',
                          'is_approved': False,
                          'rating_numerator': 0,
                          'rating_denominator': 0,
                          'is_new': False,
                          'is_popular': False,
                          'code_source': {
                              'type': 'bzr',
                              'location':
                                  'lp:~jrandom/charms/warty/name2/trunk',
                              'revision': '13',
                              'revisions': [{
                                  'revno': 32,
                                  'message': 'Hi',
                                  'date': '1998-11-30T00:00:00',
                                  'authors': [
                                      {
                                          'name': 'Baz',
                                          'email': 'baz@example.org'
                                      },
                                  ],
                              }],
                              'last_log': 'message2',
                              'bugs_link':
                              'https://bugs.launchpad.net/charms/'
                              'warty/+source/name2'
                          },
                          'relations': {
                              'provides': provides2,
                              'requires': requires2,

                          },
                          'options': options2,
                          'files': ['README.md', 'hooks/config-changed'],
                          'tested_providers': [],
                          'downloads_in_past_30_days': 0,
                          'is_subordinate': True,
                          }, result[1])
        self.assertNotEqual(one_id, two_id)
        self.assertNotEqual(one_url, two_url)

    def get_charms(self, *args, **kwargs):
        return self.get_response('charms', *args, **kwargs).json_body['result']

    def _test_filter(self, filter_str, make_charm=None):
        if make_charm is None:
            def make_charm(value, warty=False):
                kwargs = {filter_str: value}
                if filter_str != 'series' and warty:
                    kwargs['series'] = 'warty'
                return self.makeCharm(**kwargs)
        one_id, one = make_charm('common')
        self.index_client.index_charm(one)
        two_id, two = make_charm('common', warty=True)
        self.index_client.index_charm(two)
        three_id, three = make_charm('different')
        self.index_client.index_charm(three)
        four_id, four = make_charm('different-again')
        self.index_client.index_charm(four)
        result = self.get_charms(**{filter_str: 'common'})
        self.assertEqual(2, len(result))
        self.assertEqual(set([one['store_url'], two['store_url']]),
                         set(charm['url'] for charm in result))
        result = self.get_charms(**{filter_str: 'different'})
        self.assertEqual(set([three['store_url']]),
                         set(charm['url'] for charm in result))
        self.assertEqual(1, len(result))
        result = self.get_charms(**{filter_str: ['different', 'common']})
        self.assertEqual(3, len(result))

    def test_charms_name_filter(self):
        """Charms can be filtered by name."""
        self._test_filter('name')

    def test_charms_series_filter(self):
        """Charms can be filtered by series."""
        self._test_filter('series')

    def test_charms_owner_filter(self):
        """Charms can be filtered by owner."""
        self._test_filter('owner')

    def test_charms_provides_filter(self):
        """Charms can be filtered by provided interfaces."""
        def make_charm(value, warty=False):
            provides = {'foo': {'interface': value}}
            return self.makeCharm(provides=provides)
        self._test_filter('provides', make_charm)

    def test_charms_requires_filter(self):
        """Charms can be filtered by required interfaces."""
        def make_charm(value, warty=False):
            requires = {'foo': {'interface': value}}
            return self.makeCharm(requires=requires)
        self._test_filter('requires', make_charm)

    def test_charms_type_filter(self):
        """Filtering by type works as well as possible.

        Charms are considered 'approved' if owned by charmers, 'community'
        otherwise.  'environment' causes no charms to be returned.
        """
        self.makeCharm()
        self.makeCharm(owner='steve')
        result = self.get_charms(type_='community')
        self.assertEqual(['steve'], [charm['owner'] for charm in result])
        result = self.get_charms(type_='approved')
        self.assertEqual(['charmers'], [charm['owner'] for charm in result])
        result = self.get_charms(type_='environment')
        self.assertEqual([], [charm['owner'] for charm in result])
        response = self.get_response('charms', type_='foo')
        self.assertEqual(406, response.status_code)
        self.assertEqual({'type': 'unsupported_value', 'parameter': 'type',
                          'value': 'foo'}, response.json_body)
        self.assertEqual([], [charm['owner'] for charm in result])
        result = self.get_charms(type_=['approved', 'community'])
        self.assertEqual(['charmers', 'steve'],
                         [charm['owner'] for charm in result])

    def test_charms_accepts_provider(self):
        """Provider is accepted (but ignored)."""
        self.index_client.wait_for_startup()
        response = self.get_response('charms', provider='asdf')
        self.assertEqual(200, response.status_code)

    def test_charms_accepts_scope(self):
        """Scope is accepted (but ignored)."""
        self.index_client.wait_for_startup()
        response = self.get_response('charms', scope='asdf')
        self.assertEqual(200, response.status_code)

    def test_charms_accepts_category(self):
        """Category is accepted (but ignored)."""
        self.index_client.wait_for_startup()
        response = self.get_response('charms', category='asdf')
        self.assertEqual(200, response.status_code)

    def test_charms_text_search(self):
        abc_charm = self.makeCharm(summary='abc def')[1]
        self.makeCharm(summary='ghi jkl')[1]
        response = self.get_response('charms', text='def')
        self.assertEqual(200, response.status_code)
        charms = response.json_body['result']
        self.assertEqual(1, len(charms))
        self.assertEqual(API0._get_api_id(abc_charm), charms[0]['id'])

    def test_charms_text_search_respects_filters(self):
        official = self.makeCharm(summary='abc def')[1]
        self.makeCharm()
        self.makeCharm(summary='abc def', owner='steve')
        self.makeCharm(owner='steve')
        result = self.get_charms(type_=['approved'], text='def')
        self.assertEqual([API0._get_api_id(official)],
                         [charm['id'] for charm in result])

    def test_charms_limit(self):
        """Limit provides a maximum for charm numbers."""
        self.index_client.wait_for_startup()
        response = self.get_response('charms', limit='6')
        self.assertEqual(0, len(response.json_body['result']))
        self.makeCharm()
        self.makeCharm(owner='jrandom')
        response = self.get_response('charms', limit='6')
        self.assertEqual(2, len(response.json_body['result']))
        response = self.get_response('charms', limit='1')
        self.assertEqual(1, len(response.json_body['result']))
        response = self.get_response('charms', limit='0')
        self.assertEqual(0, len(response.json_body['result']))
        response = self.get_response('charms', limit='-1')
        self.assertEqual(406, response.status_code)
        self.assertEqual({
                         'type': 'negative_value',
                         'parameter': 'limit',
                         }, response.json_body)


class TestAPI0(APITestBase):

    def make_charm(self, *args, **kwargs):
        """Make a charm formatted for API0"""
        charm = factory.makeCharm(self.db, *args, **kwargs)[1]
        return API0._format_charm(charm)

    def test_unknown_endpoint(self):
        """Unknown endpoints produce a 404."""
        with self.assertRaises(HTTPNotFound):
            self.get_response('tharms')

    def test_underscore_endpoint(self):
        """Underscores in endpoints do not match."""
        with self.assertRaises(HTTPNotFound):
            self.get_response('__call__')

    def test_slash_remainder(self):
        """Endpoint urls should not be terminated with a slash."""
        with self.assertRaises(HTTPNotFound):
            self.get_response('charms', '/')

    def test_unsupported_parameter(self):
        """Unsupported parameters cause a 406."""
        response = self.get_response('charms', kermit='6')
        self.assertEqual(406, response.status_code)
        self.assertEqual({
                         'type': 'parameter_not_supported',
                         'parameter': 'kermit',
                         }, response.json_body)

    def test_format_charm_no_requires(self):
        """A missing 'requires' field is handled gracefully."""
        charmid, charm = factory.makeCharm(self.db)
        del charm['requires']
        formatted = API0._format_charm(charm)
        self.assertEqual({}, formatted['relations']['requires'])

    def test_format_charm_no_provides(self):
        """A missing 'provides' field is handled gracefully."""
        charmid, charm = factory.makeCharm(self.db)
        del charm['provides']
        formatted = API0._format_charm(charm)
        self.assertEqual({}, formatted['relations']['provides'])

    def test_format_charm_no_config(self):
        """A missing 'config' field is handled gracefully."""
        charmid, charm = factory.makeCharm(self.db)
        del charm['config']
        formatted = API0._format_charm(charm)
        self.assertEqual({}, formatted['options'])

    def test_format_charm_no_maintainer(self):
        """A missing 'maintainer' field is handled gracefully."""
        charmid, charm = factory.makeCharm(self.db)
        del charm['maintainer']
        formatted = API0._format_charm(charm)
        self.assertEqual({'name': '', 'email': ''}, formatted['maintainer'])

    def test_format_charm_adds_icon(self):
        """If a charm provides an icon, it is added to the data returned
        by the API.
        """
        charmid, charm = factory.makeCharm(self.db)
        # The test data does not contain an icon.
        self.assertFalse('icon' in charm)
        formatted = API0._format_charm(charm)
        self.assertFalse('icon' in formatted)
        charm['icon'] = 'fake'
        formatted = API0._format_charm(charm)
        self.assertTrue('icon' in formatted)

    def test_format_charm_includes_revisions(self):
        changes = [factory.makeChange(message='Ahoy!', revno=25,
                   created=1365416730.25)]
        charm = factory.get_charm_json(changes=changes)
        revisions = API0._format_charm(charm)['code_source']['revisions']
        self.assertItemsEqual(['message', 'revno', 'date', 'authors'],
                              revisions[0].keys())
        self.assertEqual('Ahoy!', revisions[0]['message'])
        self.assertEqual(25, revisions[0]['revno'])
        self.assertEqual('2013-04-08T10:25:30', revisions[0]['date'])
        self.assertEqual([
            {'name': 'Foo', 'email': 'foo@example.com'},
            {'name': 'Bar', 'email': 'bar@example.com'},
        ], revisions[0]['authors'])

    def test_charm(self):
        """Retrieving a charm works."""
        charm = self.make_charm()
        response = self.get_response(
            'charm', remainder=charm['id'])
        self.assertEqual(200, response.status_code)
        self.assertEqual(charm, response.json_body)
        response = self.get_response('charm', remainder='foo')
        self.assertEqual(404, response.status_code)
        self.assertEqual({'type': 'no_such_charm', 'charm_id': 'foo'},
                         response.json_body)

    def test_charm_wrong_revision(self):
        """Retrieving a charm works even if revision is wrong."""
        charm = self.make_charm()
        stale_charm_id = charm['id'] + '37'
        response = self.get_response('charm', remainder=stale_charm_id)
        self.assertEqual(200, response.status_code)
        self.assertEqual(charm, response.json_body)

    def test_charm_invalid_charmid(self):
        """Retrieving a charm 404s if tilde is missing."""
        charm = self.make_charm(owner='steve')
        self.assertEqual('~', charm['id'][0])
        response = self.get_response('charm', remainder=charm['id'][1:])
        self.assertEqual(404, response.status_code)

    def test_charm_url_form(self):
        """The form of the URL is not permitted to vary."""
        charm = self.make_charm()
        self.assertEqual(1, charm['id'].count('/'))
        wrong_id = '~charmers/' + charm['id']
        response = self.get_response('charm', remainder=wrong_id)
        self.assertEqual(404, response.status_code)

    def test_parse_charm_id(self):
        result = API0._parse_charm_id('~jrandom/foo/bar-23')
        self.assertEqual(('jrandom', 'foo', 'bar', '23'), result)
        result = API0._parse_charm_id('~jdoe/bar/foo-33-33')
        self.assertEqual(('jdoe', 'bar', 'foo-33', '33'), result)
        with self.assertRaises(ValueError):
            API0._parse_charm_id('jdoe/bar/foo-33-33')
        result = API0._parse_charm_id('bar/foo-33-33')
        self.assertEqual(('charmers', 'bar', 'foo-33', '33'), result)

    def test_parse_charm_id_unquotes(self):
        result = API0._parse_charm_id('~jra%3cndom/f%2eoo/ba%2fr%2d23%35')
        self.assertEqual(('jra<ndom', 'f.oo', 'ba/r', '235'), result)

    def test_extract_charm_id(self):
        charm_id, trailing = API0._extract_charm_id('/~steve/foo/bar')
        self.assertEqual('~steve/foo/bar', charm_id)
        self.assertIs(None, trailing)
        charm_id, trailing = API0._extract_charm_id(
            '/~steve/foo/bar/file/baz')
        self.assertEqual('~steve/foo/bar', charm_id)
        self.assertEqual('file/baz', trailing)
        charm_id, trailing = API0._extract_charm_id('/foo/bar/file/baz')
        self.assertEqual('foo/bar', charm_id)
        self.assertEqual('file/baz', trailing)

    def test_file(self):
        charm = factory.makeCharm(self.db)[1]
        charm_file = factory.make_charm_file(self.db, charm, 'README')
        charm_id = API0._get_api_id(charm)
        response = self.get_response(
            'charm', remainder='/' + charm_id + '/file/README')
        self.assertEqual(200, response.status_code)
        self.assertEqual(charm_file.read(), response.body)
        self.assertIn(
            ("Access-Control-Allow-Origin", "*"),
            response.headerlist)
        self.assertIn(
            ("Access-Control-Allow-Headers", "X-Requested-With"),
            response.headerlist)

    def test_file_missing(self):
        charm = factory.makeCharm(self.db)[1]
        charm_id = API0._get_api_id(charm)
        path = '/' + charm_id + '/file/subdir/README'
        response = self.get_response('charm', path)
        self.assertEqual(404, response.status_code)
        self.assertEqual({'type': 'no_such_file', 'path': 'subdir/README'},
                         response.json_body)

    def test_unknown_sub_endpoint(self):
        charm = factory.makeCharm(self.db)[1]
        charm_id = API0._get_api_id(charm)
        with self.assertRaises(HTTPNotFound):
            self.get_response('charm', '/' + charm_id + '/random')

    def test_format_category(self):
        """Formatting should return the data in the expected format.

        The expected format happens to be the current db format minus the _id,
        so formatting is trivial.
        """
        category = factory.makeQAQuestionGroup()
        category['questions'] = category['questions'][:1]
        # formatting should omit the Mongo ID.
        category['_id'] = object()
        self.assertEqual({
            'name': u'reliable',
            'description': u'Reliable',
            'questions': [{
                'id': hashlib.md5('AWS').hexdigest(),
                'description': u'AWS',
                'points': 1,
                'extended_description': u'',
            }],
        }, API0._format_category(category))

    def test_charm_qa(self):
        """"Ensure expected values are returned."""
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)
        charm = factory.makeCharm(self.db)[1]
        charm['qa'] = {
            'reliable': {
                '4847e034bb0a55fcbc8a3380d6a3ab80': '1',
                'a60b40bce263c47559d32435a89756fe': '0',
                '7ff6cc997b141399bcdf2c338fb4031c': '1',
                '9a736d9ab5ca1e16ddd942e7a62b8820': '0',
                '51dceb2909262166283ac17f2af504e7': '1',
                '9659919b570c6ef9a25df7261375e6bb': '',
            },
        }
        self.db.charms.save(charm)
        charm_id = API0._get_api_id(charm)
        response = self.get_response('charm', '/' + charm_id + '/qa')
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'result': {'questions': [API0._format_category(category)]},
            'scores': charm['qa'],
        }, response.json_body)

    def test_filter_scores(self):
        """Scores for unknown categories and questions are filtered out."""
        self.db.qa.insert(factory.makeQAQuestionGroup())
        category2 = factory.makeQAQuestionGroup()
        category2['name'] = 'useful'
        self.db.qa.insert(category2)
        charm = factory.makeCharm(self.db)[1]
        charm['qa'] = {
            'reliable': {
                '4847e034bb0a55fcbc8a3380d6a3ab80': '1',
                'bar': '1',
            },
            'foo': {},
        }
        self.db.charms.save(charm)
        charm_id = API0._get_api_id(charm)
        response = self.get_response('charm', '/' + charm_id + '/qa')
        self.assertEqual({
            'reliable': {'4847e034bb0a55fcbc8a3380d6a3ab80': '1'}},
            response.json_body['scores'])

    def test_charm_qa_no_answers(self):
        """"Handle charms that have no answers."""
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)
        charm = factory.makeCharm(self.db)[1]
        charm_id = API0._get_api_id(charm)
        response = self.get_response('charm', '/' + charm_id + '/qa')
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'result': {'questions': [API0._format_category(category)]},
            'scores': None,
        }, response.json_body)

    def test_sidebar_editorial(self):
        """A json dump of editorial charms is available."""
        res = self.get_response('sidebar_editorial')
        data = res.json_body
        self.assertIn('result', data)
        self.assertIn('new', data['result'])
        self.assertIn('slider', data['result'])
