# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from charmworld.testing import factory
from charmworld.testing import login_request
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.views.charms import charm_collection
from charmworld.views.charms import quality_edit


class TestCharms(ViewTestBase):

    def test_charm_collection_basic(self):
        """Simple sanity check."""
        one_id, one = factory.makeCharm(self.db)
        two_id, two = factory.makeCharm(self.db)

        response = charm_collection(self.getRequest())
        charm_names = set(charm['name'] for charm in response['charms'])
        expected = set((one['name'], two['name']))
        self.assertEqual(expected, charm_names)

    def test_charm_collection_sub_filter(self):
        """With the subordinate filter we limit out the non-sub charms."""
        one_id, one = factory.makeCharm(self.db, subordinate=True)
        two_id, two = factory.makeCharm(self.db)

        request = self.getRequest()
        request.GET['sub'] = 'true'
        response = charm_collection(request)
        charms = response['charms']
        self.assertEqual(1, len(charms))
        self.assertEqual(one['name'], charms[0]['name'])


class TestQAForm(ViewTestBase):

    def setUp(self):
        """Need to load the question data."""
        super(TestQAForm, self).setUp()
        category = factory.makeQAQuestionGroup()
        self.db.qa.insert(category)

    def test_form_renders(self):
        """Sanity check that the form renders without errors."""
        cid, charm = factory.makeCharm(self.db)
        request = self.getRequest()
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }
        resp = quality_edit(request)

        self.assertIn('form', resp)
        self.assertIn('label', resp['form'])

    def test_charm_saves_data(self):
        """Verify good data causes the view to store qa data.

        This also verifies our schema a little bit as the fields are turned
        from strings in POST to int's per the schema to match the original
        factory generated data.

        """
        cid, charm = factory.makeCharm(self.db)
        post_data = factory.makeQAAssessment()
        # Add the submit button to trigger the right path through the view.
        post_data['submit'] = 'submit'
        request = self.getRequest(post=post_data)
        request.matchdict = {
            'charm': charm['name'],
            'series': charm['series'],
        }

        quality_edit(request)

        # Now fetching that charm should get us a charm with a qa attribute
        # matching the data submitted.
        found = self.db.charms.find_one({
            '_id': cid
        })

        self.assertIn('qa', found)
        # Remove the submit button we added since it's not stored into the
        # document.
        del post_data['submit']

        for key in post_data:
            # Ignore any form keys that start with __
            if key.startswith('__'):
                continue

            # Ignore the deform fields that make up the nested structure for
            # the qa questions.
            if key.startswith('deform'):
                continue

            self.assertEqual(
                post_data[key],
                found['qa'][key])

        # The mapped/nested data needs to exist in the updated charm data.
        self.assertIn('reliable', found['qa'])


class TestQAFormFunc(WebTestBase):
    """Functional tests for the QA Form Views."""

    def test_form_403s_for_non_charmers(self):
        """The QA form is not permitted if you're not a charmer."""
        with login_request(self.app, groups=[]) as resp:
            cid, charm = factory.makeCharm(self.db)
            # The status force the check against the status code.
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['series'], charm['name']),
                status=403)
            self.assertEqual(403, resp.status_code)

    def test_form_secured_from_users(self):
        """The qa form is only allowed to be edited/submitted by charmers."""
        with login_request(self.app, groups=['charmers']) as resp:
            cid, charm = factory.makeCharm(self.db)
            resp = resp.goto(
                'charms/%s/%s/qa/edit' % (charm['name'],
                charm['series']),
                status=200)
            self.assertEqual(resp.status_code, 200)


class FakeReadme:
    """charmworld.views.charms._charm_readme() tries to find a readme file
    in the local copy the source code branch, but we don't have real
    branches available for for test data. Return fake data instead.
    """
    def __call__(self, charm):
        return '', "A README text"
